"""Configuration document.
This module contains class definitions and functions to read from parsed file.
"""

import re
from enum import Enum

from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, TypeVar, Union
from typing_extensions import TypedDict

import ukosnik.docent as doc


# Errors
class ContextualError(doc.ReadError):
    """Error adding context to a child error that was raised."""


class ValidationError(doc.ReadError):
    """Raised when a value doesn't pass validation (e.g. pattern restriction)."""


class InvalidOptionTypeError(doc.ReadError):
    """Raised when an option type cannot be parsed (e.g. has a typo)."""


# Constants
NAME_PATTERN = re.compile(r"^[\w-]{1,32}$")


class MetaType(Enum):
    """Meta class type. Used to print more context in errors."""

    COMMAND = "command"
    OPTION = "option"


class CommandOptionType(Enum):
    """docs/interactions/slash-commands#applicationcommandoptiontype"""

    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9

    @staticmethod
    def from_str(name: str):
        """Parses an option type from string. Case do not matter.
        Aliases: SUBCOMMAND/SUB_COMMAND, SUBCOMMAND_GROUP/SUB_COMMAND_GROUP.
        """

        key = name.upper()
        if key.startswith("SUBCOMMAND"):
            key = "SUB_COMMAND" + key[len("SUBCOMMAND"):]
        if key in CommandOptionType.__dict__:
            return CommandOptionType[key]
        raise InvalidOptionTypeError(f"'{name.upper()}' is not a known command option type.")


# Typed dicts
class Meta(TypedDict):
    """Shared meta fields for other typed dicts. Related to `MetaType`."""

    name: str
    description: str


class Command(Meta):
    """docs/interactions/slash-commands#applicationcommand
    Fields `id` and `application_id` are present only when fetched from the API.
    """

    id: int
    application_id: int
    options: Optional[List["CommandOption"]]
    default_permission: Optional[bool]


class CommandOption(Meta):
    """docs/interactions/slash-commands#applicationcommandoption"""

    type: int
    required: Optional[bool]
    choices: Optional[List["CommandOptionChoice"]]
    options: Optional[List["CommandOption"]]


class CommandOptionChoice(TypedDict):
    """docs/interactions/slash-commands#applicationcommandoptionchoice"""

    name: str
    value: Union[str, int]


# Classes
@dataclass
class Document:
    """Main configuration document.
    Attributes' dictionnaries can be sent raw to the Discord API. The Typed Dicts are
    conform to types described on https://discord.com/developers/docs/interactions/slash-commands.

    Attributes:
        commands — command list, already validated, command dicts can be sent raw to Discord API
    """

    commands: List[Command]


# Readers
def read(doc_dict: Dict[str, Any]) -> Document:
    """Reads a dictionary representing a document into a document.
    The dict may come from parsed yaml, json, toml, and whatnot.
    """

    try:
        commands = doc.read(doc_dict, "commands", doc.with_default(read_commands, []))
    except doc.ReadError as err:
        raise ContextualError(f"While reading commands: {err}") from err
    return Document(commands)


def read_commands(data: Any) -> List[Command]:
    """Reads a list of commands from raw data."""
    def __fn(doc_command: Dict[str, Any]) -> Command:
        name = doc.read(doc_command, "name", doc.typed(str))
        description = doc.read(doc_command, "description", doc.typed(str))
        validate_meta(MetaType.COMMAND, name, description)
        command = {
            "name": name,
            "description": description,
        }
        doc.read(doc_command, "options", doc.with_default(read_options, None), to=command)
        doc.read(
            doc_command,
            "default-permission",
            doc.typed(bool, optional=True),
            to=command,
            to_key="default_permission",
        )
        return command  # type: ignore

    return __read_list_or_keyed("Command", data, __fn)


def read_options(data: Any) -> List[CommandOption]:
    """Reads a list of command options from raw data"""
    def __fn(doc_option: Dict[str, Any]) -> CommandOption:
        name = doc.read(doc_option, "name", doc.typed(str))
        description = doc.read(doc_option, "description", doc.typed(str))
        validate_meta(MetaType.OPTION, name, description)
        kind_key = doc.read(doc_option, "type", doc.typed(Union[str, int]))
        try:
            kind = CommandOptionType.from_str(kind_key) if isinstance(kind_key, str) else CommandOptionType(kind_key)
        except ValueError as err:
            raise InvalidOptionTypeError(" ".join(
                [f"{kind_key} is not a valid command option type.",
                 "It must be between 1 and 9 (both inclusive)."])) from err
        option = {
            "name": name,
            "description": description,
            "type": kind.value,
        }
        doc.read(doc_option, "required", doc.typed(bool, optional=True), to=option)
        if "choices" in doc_option and isinstance(doc_option["choices"], list):
            choices = []
            for doc_choice in doc_option["choices"]:
                if isinstance(doc_choice, (str, int)):
                    doc_choice = {"name": doc_choice}
                if not isinstance(doc_choice, dict):
                    raise doc.ValueTypeError(f"Choice of unexpected type '{type(doc_choice).__name__}'")
                choice_name = doc.read(doc_choice, "name", doc.typed(str))
                validate_length("Choice name", choice_name)
                choices.append({
                    "name": choice_name,
                    "value": doc.read(doc_choice, ["value", "name"], doc.typed(Union[str, int]))
                })
            option["choices"] = choices
        doc.read(doc_option, "options", doc.with_default(read_options, None), to=option)
        return option  # type: ignore

    return __read_list_or_keyed("Option", data, __fn)


T = TypeVar("T")


def __read_list_or_keyed(kind: str, data: Any, fn: Callable[[Dict[str, Any]], T]) -> List[T]:
    """DRY function to read either a list or a dictionnary with value's name being the key."""
    if isinstance(data, dict):
        doc_values = []
        for name, doc_value in data.items():
            if not isinstance(doc_value, dict):
                raise doc.ValueTypeError(f"{kind} '{name}' is not an object.")
            if not "name" in doc_value:
                doc_value["name"] = name
            doc_values.append(doc_value)
    elif isinstance(data, list):
        doc_values = data
    else:
        raise doc.ValueTypeError(f"{kind}s must either be an array or an object with option names as keys.")

    options = []
    for doc_value in doc_values:
        try:
            options.append(fn(doc_value))
        except doc.ReadError as err:
            name = doc_value["name"] if "name" in doc_value else "unnamed"
            raise ContextualError(f"'{name}' > {err}") from err
    return options


def validate_meta(kind: MetaType, name: str, description: str):
    """Validates meta fields, raise an exception if a validation fails."""
    if not NAME_PATTERN.match(name):
        raise ValidationError(f"Command name '{name}' must be alphanumeric and less than 32 chars long.")
    if description is None:
        raise ValidationError(f"Description is required for {kind.value} '{name}'.")
    validate_length("Description", description)


def validate_length(kind: str, value: str, _min: int = 1, _max: int = 100):
    """Validates length of a value."""
    if len(value) < _min:
        raise ValidationError(f"{kind} '{value}' must be at least {_min} chars long.")
    if len(value) > _max:
        raise ValidationError(f"{kind} '{value}' must be strictly less than {_max + 1} chars long.")



"""Ukosnik - Manage Discord slash commands.
Main entry file for command-line usage.
"""

import argparse

import os
import sys
import traceback

from ukosnik import __version__
from ukosnik.docent import ReadError
from ukosnik.http import Client, CommandManager, HTTPRequestException
import ukosnik.document as ukodoc

cli = argparse.ArgumentParser(description="Manage slash commands easily for your Discord bot")
cli.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
subparsers = cli.add_subparsers(dest="subcommand")

# args
cli.add_argument("--token", help="bot token (default: ENV['DISCORD_TOKEN'])")
cli.add_argument("--app", type=int, help="application id (default: fetched automatically)")
cli.add_argument("--colored", type=bool, default=True, help="whether some messages should be printed with colors")

# subcommands
cli_update = subparsers.add_parser("update", help="Update application's slash commands")
cli_update.add_argument("config", help="config file")
cli_update.add_argument("-d", action="store_true", help="delete existing commands not present in config")

cli_clear = subparsers.add_parser("clear", help="Clear application's slash commands")

for parser in (cli_update, cli_clear):
    parser.add_argument("--guild", type=int, help="use this guild's scope (default: application's scope)")

app_args = cli.parse_args()
if not app_args.subcommand:
    cli.print_help()
    sys.exit(0)


def colored(color: str, *args, **kwargs):
    if app_args.colored and len(args) > 0:
        print(f"{color}{args[0]}", *args[1:], "\033[0m", **kwargs)
    else:
        print(*args, **kwargs)


def fail(*args, **kwargs):
    if "file" not in kwargs:
        kwargs["file"] = sys.stderr
    colored("\033[91m", *args, **kwargs)
    sys.exit(1)


def info(*args, **kwargs):
    colored("\033[94m", "I:", *args, **kwargs)


def ok(*args, **kwargs):
    colored("\033[92m", "OK:", *args, **kwargs)


token = app_args.token or os.environ.get("DISCORD_TOKEN")
if token is None:
    fail(
        "Bot token is required! You can pass it through the DISCORD_TOKEN",
        "env variable or the '--token' argument.",
        file=sys.stderr,
    )
if " " not in token:
    token = f"Bot {token}"

if app_args.subcommand == "update":
    try:
        with open(app_args.config, "r") as config_file:
            try:
                _, ext = os.path.splitext(app_args.config)
                if ext in (".yml", ".yaml"):
                    from yaml import full_load as read_file
                elif ext == ".json":
                    from json import load as read_file
                else:
                    raise Exception("Unsupported file type '%s'." % ext)
                doc_dict = read_file(config_file)
            except Exception:
                traceback.print_exc()
                fail("Could not read config file.")

            try:
                document = ukodoc.read(doc_dict)
                del doc_dict
            except ReadError as err:
                fail(f"Could not parse configuration file.\nError: {err}")
    except IsADirectoryError:
        fail("Configuration file cannot be a directory.")
    except FileNotFoundError as err:
        fail(f"Configuration file not found. ({err})")

    if len(document.commands) > 0:
        info(f"Loaded {len(document.commands)} command(s).")
    else:
        info("No commands were loaded (empty 'commands' list).")

try:
    client = Client(token)
    application_id = app_args.app or client.get_application_id()
    commands_manager = CommandManager(client, application_id, app_args.guild)

    info(f"Using application ID {application_id}.")
    if commands_manager.guild_id:
        info(f"Using guild ID {commands_manager.guild_id}.")

    existing_commands = commands_manager.fetch_commands()
    info(f"Fetched {len(existing_commands)} existing commands.")

    if app_args.subcommand == "update":
        # doc is defined but ls struggles to know it
        # pylint: disable=self-assigning-variable
        document = document  # type: ignore

        existing_command_names = set(map(lambda command: command["name"], existing_commands))
        command_names = set(map(lambda command: command["name"], document.commands))

        # why tf does pylint think step should be a constant
        step = 1  # pylint: disable=C0103
        print(f"[{step}] Updating commands...")
        if len(document.commands) == 0:
            info("No commands to insert/update.")
        for command in document.commands:
            command_name = command["name"]
            try:
                command_id = commands_manager.upsert_command(command)["id"]
                VERB = "Updated" if command_name in existing_command_names else "Registered"
                ok(f"{VERB} command '{command_name}' ({command_id}) successfully.")
            except HTTPRequestException as err:
                VERB = "update" if command_name in existing_command_names else "register"
                fail(f"Failed to {VERB} command {command_name}.\nError: {err}")

        step += 1
        if app_args.d:
            print(f"[{step}] Delete flag detected, deleting former commmands...")
            former_commands = [cmd for cmd in existing_commands if cmd["name"] not in command_names]
            if len(former_commands) == 0:
                info("No former commands to delete.")
            for command in former_commands:
                command_name = command["name"]
                try:
                    commands_manager.delete_command(command["id"])
                    ok(f"Deleted command '{command_name}' ({command['id']}) sucessfully.")
                except HTTPRequestException as err:
                    fail(f"Failed to delete command {command_name}.\nError: {err}")
        print("Slash commands updated successfully.")
    elif app_args.subcommand == "clear":
        print("Clearing all slash commands...")
        for command in existing_commands:
            command_name = command["name"]
            try:
                commands_manager.delete_command(command["id"])
                ok(f"Deleted command '{command_name}' ({command['id']}) sucessfully.")
            except HTTPRequestException as err:
                fail(f"Failed to delete command {command_name}.\nError: {err}")
        print("Cleared slash commands successfully.")

except HTTPRequestException as err:
    fail(f"An error was raised while performing a request.\nError: {err}")



"""Small wrapper around Discord endpoints.
Requests should not be handled in any other module.
"""

from typing import Any, Dict, List, Optional
import requests
from requests.models import Response

from ukosnik import __version__
from ukosnik.document import Command


class HTTPRequestException(Exception):
    """Raised when something failed while performing a request."""


class Client:
    """Small Discord HTTP client."""
    def __init__(
        self,
        token: str,
        base_url: str = "https://discord.com/api/v9",
    ):
        assert token is not None, "Token required"
        assert base_url is not None and base_url.startswith("https://"), "Base URL must use HTTPS"

        self.base_url = base_url.rstrip('/')
        self.headers = {
            "Authorization": token,
            "User-Agent": f"https://github.com/Unoqwy/ukosnik (v{__version__})",
            "Content-Type": "application/json"
        }

    @staticmethod
    def __handle(response: Response) -> Response:
        if response.status_code == 401:
            raise HTTPRequestException(
                "The bot token does not grant access to the required API endpoints.",
                "Possible reasons include: expired token, ungranted scopes.",
            )
        return response

    def get(self, endpoint: str) -> Response:
        return self.__handle(requests.get(f"{self.base_url}/{endpoint}", headers=self.headers))

    def post(self, endpoint: str, data: Dict[str, Any]) -> Response:
        return self.__handle(requests.post(f"{self.base_url}/{endpoint}", headers=self.headers, json=data))

    def delete(self, endpoint: str) -> Response:
        return self.__handle(requests.delete(f"{self.base_url}/{endpoint}", headers=self.headers))

    def get_application_id(self) -> int:
        return self.get("oauth2/applications/@me").json()["id"]


class CommandManager:
    """Contextual wrapper arround command endpoints."""
    def __init__(self, client: Client, application_id: int, guild_id: Optional[int]):
        self.client = client
        self.application_id = application_id
        self.guild_id = guild_id

        base = f"applications/{self.application_id}"
        if self.guild_id is not None:
            base = f"{base}/guilds/{self.guild_id}"
        self.endpoint = f"{base}/commands"

    def command_endpoint(self, command_id: int) -> str:
        return f"{self.endpoint}/{command_id}"

    def fetch_commands(self) -> List[Command]:
        return self.client.get(self.endpoint).json()

    def upsert_command(self, command: Command) -> Command:
        return self.client.post(self.endpoint, command).json()  # type: ignore

    def delete_command(self, command_id: int):
        self.client.delete(self.command_endpoint(command_id))



"""Docent, the guiding module.
Contains helper functions to read from a dict.

The functions in this module should most likely not be used externally.
"""

from inspect import signature
from types import FunctionType
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union, cast


# Errors
class ReadError(Exception):
    """Base error for when a given dict cannot be read into a document."""


class ValueTypeError(ReadError):
    """Raised when a value has an unexpected type."""


class MissingValueError(ReadError):
    """Raised when a required value is missing."""


# Generics and type aliases
T = TypeVar("T")
DocDict = Dict[str, Any]
ReadFn = Callable[[Optional[str], Any], T]  # closure taking key, value
SimpleFn = Callable[[Any], T]  # closure taking only value


# Read
def read(doc_dict: DocDict,
         keys: Union[str, List[str]],
         fn: ReadFn[T],
         to: Optional[DocDict] = None,
         to_key: Optional[str] = None) -> T:
    assert isinstance(fn, FunctionType), "fn must be a valid function"
    if isinstance(keys, str):
        keys = [keys]

    key, value = __read(doc_dict, keys, fn)
    if to is not None:
        to_key = to_key if to_key is not None else key
        if to_key is not None and value is not None:
            to[to_key] = value
    return value


def __read(doc_dict: DocDict, keys: List[str], fn: ReadFn[T]) -> Tuple[Optional[str], T]:
    for key in keys:
        if key in doc_dict:
            return key, fn(key, doc_dict[key])
    key = keys[0] if len(keys) > 0 else None
    return key, fn(key, None)


def with_default(parent: Union[ReadFn[T], SimpleFn[T]], default: T) -> ReadFn[T]:
    if len(signature(parent).parameters) == 1:
        fn = lambda _, value: parent(value)  # type: ignore
    else:
        fn = cast(ReadFn[T], parent)
    return lambda key, value: fn(key, value) if value is not None else default


def typed(type_hint: Type, optional: bool = False) -> ReadFn[Any]:
    def __fn(key: Optional[str], value: Any) -> Any:
        if optional and value is None:
            return value
        if value is None:
            raise ValueTypeError(f"'{key}' is required yet missing.")

        args = type_hint.__args__ if hasattr(type_hint, "__args__") else type_hint
        if not isinstance(value, args):
            raise ValueTypeError(" ".join([
                f"Expected value from '{key}' to be of type '{type_hint.__name__}'",
                f"but is it of type '{type(value).__name__}' instead."
            ]))
        return value

    return __fn
